from typing import Callable, Dict, List

from pydantic import Field
from typing_extensions import Annotated

from lmfunctions.base import Base
from lmfunctions.handler import (
    ExceptionRaiseHandler,
    LoggingHandler,
    PanelPrintHandler,
    PrintHandler,
)

StandardHandler = Annotated[
    LoggingHandler | PanelPrintHandler | PrintHandler | ExceptionRaiseHandler,
    Field(discriminator="name"),
]

Handler = StandardHandler | Callable


class EventManager(Base):
    """
    An EventManager defines a map between event names and a list of handler functions.
    Each handler is a callable that is executed when the corresponding event occurs.

    There are two types of handlers: standard handlers and custom handlers. Standard
    handlers are predefined functions which can be serialized, whereas custom
    handlers are arbitrary callables.
    """

    handlers: Dict[str, List[Handler]] = dict(
        exception=[ExceptionRaiseHandler(varname="exception")],
    )

    def __call__(self, event_name, **kwargs):
        [handler(**kwargs) for handler in self.handlers.get(event_name, [])]
        return None

    def __add__(self, other):
        if isinstance(other, EventManager):
            handlers = {}
            for d in (self.handlers, other.handlers):
                for key, value in d.items():
                    handlers.setdefault(key, []).extend(value)
            return EventManager(handlers=handlers)
        return NotImplemented



import json
from typing import Any, Callable, Dict, Iterator, Optional, Union

from lmfunctions.base import Base
from lmfunctions.handler import PrintHandler


class LMResponse(Base):
    """
    A class representing a response from a language model.

    The response can be initialized with a string or an iterator of strings.

    Executing __call__ on the object will iterate over tokens/characters
    in the response, optionally call a callback function for each token/character
    processed, and parse the response according to a given schema.
    """

    text: str = ""

    _unprocessed: Union[str, Iterator[str]]

    def __init__(self, unprocessed: Union[str, Iterator[str]], **kwargs):
        super().__init__(**kwargs)
        self._unprocessed = unprocessed

    def __repr__(self):
        self()
        print()
        return super().__repr__()

    def __call__(
        self,
        schema: Optional[Dict] = None,
        new_token_or_char_callback: Optional[Callable] = PrintHandler(
            varnames=["token_or_char"], end="", flush=True
        ),
        **kwargs
    ) -> Any:
        """
        Processes the response, optionally parsing it according to the given schema.
        For each token or character processed, calls the `new_token_or_char_callback` function
        if provided.

        Args:
            schema (Optional[Dict]): A JSON schema to parse the response.
            new_token_or_char_callback (Optional[Callable]): A callback function
            to be called when a new token/character is processed. Defaults to PrintHandler.

        Returns:
            Any: The processed response.

        Raises:
            JSONDecodeError: If JSON parsing is required but no valid JSON is found.
        """
        if not self.text:
            # Check if the schema defines a JSON object
            json_object = (schema is not None) and (
                schema.get("type", None) == "object"
            )
            # The response is unprocessed
            depth, in_json = 0, False
            for token_or_char in self._unprocessed:
                open_braces, closed_braces = token_or_char.count(
                    "{"
                ), token_or_char.count("}")
                if not (json_object) or in_json or open_braces:
                    self.text += token_or_char
                    # New token or character callback
                    if new_token_or_char_callback:
                        new_token_or_char_callback(
                            schema=schema,
                            json_object=json_object,
                            token_or_char=token_or_char,
                            depth=depth,
                            in_json=in_json,
                            open_braces=open_braces,
                            closed_braces=closed_braces,
                            **kwargs
                        )
                if json_object:
                    if in_json:
                        depth += open_braces - closed_braces
                        if depth == 0:
                            try:
                                return json.loads(self.text)
                            except json.JSONDecodeError:
                                self.text, in_json = "", False
                    elif open_braces:
                        depth, in_json = 1, True
        return (
            json.loads(self.text)
            if (schema and schema.get("type", None) != "string")
            else self.text
        )

    @classmethod
    def from_openai_v1(cls, response: Union[Any, Iterator[Any]]):
        """
        Creates an LMResponse object from a response in OpenAI-v1 format.
        """
        if isinstance(response, Iterator):
            response_generator = (c.choices[0].delta.content or "" for c in response)
            return cls(response_generator)
        else:
            content = response.choices[0].message.content
            return cls(content if isinstance(content, str) else "")



from typing import Any, Dict, Iterable, List, Literal, Optional, Union

from lmfunctions.base import Base
from lmfunctions.lmbackend.lmbackend import LMBackend
from lmfunctions.lmresponse import LMResponse
from lmfunctions.utils import lazy_import


class LiteLLMBackend(LMBackend, Base):

    name: Literal["litellm"] = "litellm"
    model: str = "gpt-4o-mini"
    timeout: float | None = None
    temperature: float | None = None
    top_p: float | None = None
    n: int | None = None
    stream: bool | None = True
    stream_options: Dict | None = None
    stop: str | None = None
    max_tokens: int | None = None
    presence_penalty: float | None = None
    frequency_penalty: float | None = None
    logit_bias: Dict | None = None
    user: str | None = None
    response_format: Dict | None = None
    seed: int | None = None
    tools: list[Any] | None = None
    tool_choice: str | None = None
    logprobs: bool | None = None
    top_logprobs: int | None = None
    extra_headers: Dict | None = None
    functions: List | None = None
    function_call: str | None = None
    base_url: str | None = None
    api_version: str | None = None
    api_key: str | None = None
    max_retries: int = 2
    mock_response: str | None = None

    def complete(
        self, prompt: str = "", schema: Optional[Dict] = None, **kwargs
    ) -> LMResponse:
        messages = [{"role": "user", "content": prompt}]
        response_format = dict(type="json_object") if schema else None
        return self.chat_complete(messages, response_format=response_format, **kwargs)

    def chat_complete(self, messages: List, **kwargs) -> LMResponse:
        response = self.chat_complete_openai_v1(messages, **kwargs)
        return LMResponse.from_openai_v1(response)

    def chat_complete_openai_v1(
        self, messages: List, **kwargs
    ) -> Union[Any, Iterable[Any]]:
        params = self.model_dump(exclude={"name"}) | kwargs
        lazy_import("litellm")
        import litellm

        litellm.drop_params = True
        return litellm.completion(messages=messages, **params)



import os

from lmfunctions.objectstore.objectstore import ObjectStore
from lmfunctions.utils import loadf

BASE_URL: str = os.environ.get("LMHUB_URL", "https://lmhub.steerable.co/lmhub/")


class Hub(ObjectStore):
    def push(self, obj, path, **kwargs):
        raise NotImplementedError(
            "Pushing objects is not supported yet."
        )  # pragma: no cover

    @staticmethod
    def pull(path, format="yaml", **kwargs):
        url = f"{BASE_URL}{path}.{format}"
        return loadf(url, **kwargs)



from abc import abstractmethod
from typing import Any

from lmfunctions.base import Base


class Handler(Base):

    @abstractmethod
    def __call__(self, **kwargs) -> Any:
        pass  # pragma: no cover



from lmfunctions.eventmanager import EventManager
from lmfunctions.lmbackend import LMBackend, LlamaCppBackend
from lmfunctions.retrypolicy import RetryPolicy


class default:
    backend: LMBackend = LlamaCppBackend()
    backend.generation.stop = [
        "<|end_header_id|>",
        "<|eot_id|>",
        "<|reserved_special_token",
    ]
    event_manager = EventManager()
    retry_policy = RetryPolicy()



from lmfunctions.eventmanager import EventManager
from lmfunctions.handler import PrintHandler


def tokenStream() -> EventManager:
    """
    EventManager preset that prints individual tokens or characters to the console as
    they are processed.
    """
    return EventManager(
        handlers={
            "token_or_char": [
                PrintHandler(varnames=["token_or_char"], end="", flush=True)
            ]
        }
    )



import subprocess


def cuda_check():
    gpu_info = {"cuda_available": False, "num_gpus": 0, "gpus": []}

    try:
        # Check for CUDA availability using nvidia-smi
        result = subprocess.run(
            ["nvidia-smi", "--query-gpu=name", "--format=csv,noheader"],
            capture_output=True,
            text=True,
        )

        if result.returncode != 0:
            return gpu_info

        gpu_names = result.stdout.strip().split("\n")
        gpu_info["cuda_available"] = True
        gpu_info["num_gpus"] = len(gpu_names)

        # Query GPU memory
        result = subprocess.run(
            [
                "nvidia-smi",
                "--query-gpu=memory.total,memory.free,memory.used",
                "--format=csv,noheader,nounits",
            ],
            capture_output=True,
            text=True,
        )
        memory_info = result.stdout.strip().split("\n")

        for i, line in enumerate(memory_info):
            total_memory, free_memory, used_memory = line.split(", ")
            gpu_info["gpus"].append(
                {
                    "name": gpu_names[i],
                    "total_memory": int(total_memory),
                    "free_memory": int(free_memory),
                    "used_memory": int(used_memory),
                }
            )

    except FileNotFoundError:
        pass

    return gpu_info



import gc
import json
import multiprocessing
from typing import Any, Dict, Iterator, List, Literal, Optional, Union

import huggingface_hub
from pydantic import model_validator

from lmfunctions.base import Base
from lmfunctions.lmbackend.lmbackend import LMBackend
from lmfunctions.lmresponse import LMResponse
from lmfunctions.utils import cuda_check, lazy_import


class LLamaCppGenerationParams(Base):
    """
    Parameters controlling generation with the LLamaCpp model.
    """

    max_tokens: int | None = None
    temperature: float = 0.8
    top_p: float = 0.95
    min_p: float = 0.05
    typical_p: float = 1
    logprobs: int | None = None
    stop: str | List[str] | None = []
    frequency_penalty: float = 0
    presence_penalty: float = 0
    repeat_penalty: float = 1.1
    top_k: int = 40
    stream: bool = True
    seed: int | None = None
    tfs_z: float = 1
    mirostat_mode: int = 0
    mirostat_tau: float = 5
    mirostat_eta: float = 0.1
    model: str | None = None
    logit_bias: Dict[str, float] | None = None


class LlamaCppBackend(LMBackend):
    """
    LLamaCpp model configuration.  The model is loaded lazily, only when it is accessed.
    If any of the arguments get modified, the current model is unloaded.
    """

    name: Literal["llamacpp"] = "llamacpp"
    model: str = (
        "hf://QuantFactory/Meta-Llama-3.1-8B-Instruct-GGUF/Meta-Llama-3.1-8B-Instruct.Q4_K_M.gguf"
    )
    n_gpu_layers: Optional[int] = None
    split_mode: int = 1
    main_gpu: int = 0
    tensor_split: List[float] | None = None
    vocab_only: bool = False
    use_mmap: bool = True
    use_mlock: bool = False
    kv_overrides: Dict[str, bool | int | float] | None = None
    seed: int = 4294967295
    n_ctx: int = 0
    n_batch: int = 512
    n_threads: int | None = None
    n_threads_batch: int | None = None
    rope_scaling_type: int | None = -1
    rope_freq_base: float = 0
    rope_freq_scale: float = 0
    yarn_ext_factor: float = -1
    yarn_attn_factor: float = 1
    yarn_beta_fast: float = 32
    yarn_beta_slow: float = 1
    yarn_orig_ctx: int = 0
    mul_mat_q: bool = True
    logits_all: bool = False
    embedding: bool = False
    offload_kqv: bool = True
    last_n_tokens_size: int = 64
    lora_base: str | None = None
    lora_scale: float = 1
    lora_path: str | None = None
    numa: bool = False
    chat_format: str | None = None
    verbose: bool = False
    generation: LLamaCppGenerationParams = LLamaCppGenerationParams()

    _llama: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.n_gpu_layers is None:
            gpu_info = cuda_check()
            if gpu_info["cuda_available"]:
                self.n_gpu_layers = -1
            else:
                self.n_gpu_layers = 0
        if self.n_threads is None:
            self.n_threads = multiprocessing.cpu_count() // 2

    @property
    def llama(self):
        if self._llama is None:
            lazy_import("llama_cpp", install_packages=["llama-cpp-python"])
            from llama_cpp import Llama

            model_reference = self.model
            if model_reference.startswith("hf://"):
                components = model_reference.split("://")[1].split("/")
                repo_id, filename = "/".join(components[0:2]), components[-1]
                model_path = huggingface_hub.hf_hub_download(repo_id, filename)
            else:
                model_path = model_reference
            self._llama = Llama(
                model_path=model_path,
                **self.model_dump(exclude={"name", "model", "generation"})
            )
        return self._llama

    def _unload(self):
        self._llama = None
        gc.collect()

    @model_validator(mode="after")
    def unload(self):
        if self._llama:
            self._unload()
        return self

    def complete(
        self, prompt: str = "", schema: Optional[Dict] = None, **kwargs
    ) -> LMResponse:

        if "tokenizer.chat_template" in self.llama.metadata:
            messages = [{"role": "user", "content": prompt}]
            response_format = (
                dict(type="json_object", schema=schema) if schema else None
            )
            return self.chat_complete(
                messages, response_format=response_format, **kwargs
            )
        else:
            lazy_import("llama_cpp", install_packages=["llama-cpp-python"])
            from llama_cpp.llama_chat_format import _grammar_for_json_schema

            grammar = _grammar_for_json_schema(json.dumps(schema)) if schema else None
            params = self.generation.model_dump() | kwargs
            response = self.llama.create_completion(prompt, grammar=grammar, **params)
            if isinstance(response, Iterator):
                response_generator = (c["choices"][0]["text"] or "" for c in response)
                return LMResponse(response_generator)
            else:
                content = response["choices"][0]["text"]
                return LMResponse(content if isinstance(content, str) else "")

    def chat_complete(self, messages: List, **kwargs) -> LMResponse:
        response = self.chat_complete_openai_v1(messages, **kwargs)
        return LMResponse.from_openai_v1(response)

    def chat_complete_openai_v1(
        self, messages: List, **kwargs
    ) -> Union[Any, Iterator[Any]]:
        params = self.generation.model_dump() | kwargs
        return self.llama.create_chat_completion_openai_v1(messages=messages, **params)



from abc import abstractmethod
from typing import Dict, List, Optional

from lmfunctions.base import Base
from lmfunctions.lmresponse import LMResponse


class LMBackend(Base):
    """
    A class providing methods for interacting with a language model backend,
    such as completing prompts and replying to chat messages.
    """

    @abstractmethod
    def complete(
        self, prompt: str, schema: Optional[Dict] = None, *args, **kwargs
    ) -> LMResponse:
        pass  # pragma: no cover

    @abstractmethod
    def chat_complete(self, messages: List, **kwargs) -> LMResponse:
        pass  # pragma: no cover



from lmfunctions.base.base import Base
from lmfunctions.eventmanager import (
    EventManager,
    consoleRich,
    fileLog,
    panelPrint,
    tokenStream,
)
from lmfunctions.eventmanager.managers import timeEvents
from lmfunctions.lmbackend import (
    LiteLLMBackend,
    LlamaCppBackend,
    TransformersBackend,
)
from lmfunctions.retrypolicy import RetryPolicy

from .chat import chat
from .default import default
from .lmbackend import LMBackend
from .lmfunc import LMFunc, lmdef
from .lmresponse import LMResponse

# Shortcuts
from_string = LMFunc.from_string
from_store = LMFunc.from_store
from_file = LMFunc.from_file


def complete(*args, **kwargs):
    return default.backend.complete(*args, *kwargs)


class BackendSetter:
    @staticmethod
    def llamacpp(*args, **kwargs):
        default.backend = LlamaCppBackend(*args, **kwargs)

    @staticmethod
    def transformers(*args, **kwargs):
        default.backend = TransformersBackend(*args, **kwargs)

    @staticmethod
    def litellm(*args, **kwargs):
        default.backend = LiteLLMBackend(*args, **kwargs)


set_backend = BackendSetter()


class EventManagerSetter:
    @staticmethod
    def panelprint():
        default.event_manager = panelPrint()

    @staticmethod
    def consolerich():
        default.event_manager = consoleRich()

    @staticmethod
    def filelogger(*args, **kwargs):
        default.event_manager = fileLog(*args, **kwargs)

    @staticmethod
    def tokenstream():
        default.event_manager = tokenStream()

    @staticmethod
    def time_events():
        default.event_manager = timeEvents()

    @staticmethod
    def default():
        default.event_manager = EventManager()


set_event_manager = EventManagerSetter()

__all__ = [
    "lmdef",
    "LMFunc",
    "chat",
    "from_string",
    "from_store",
    "from_file",
    "LMBackend",
    "LMResponse",
    "EventManager",
    "Base",
    "RetryPolicy",
    "set_backend",
    "set_event_manager",
    "complete",
    "default",
]



import inspect
import json
import os
import re
from types import NoneType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    ParamSpec,
    Type,
    TypeVar,
    get_type_hints,
)

from jinja2 import Template
from opentelemetry import trace
from pydantic import BaseModel, create_model
from tenacity import Retrying

from lmfunctions.base import Base
from lmfunctions.default import default
from lmfunctions.eventmanager import EventManager
from lmfunctions.lmbackend import LMBackend
from lmfunctions.retrypolicy import RetryPolicy

# from lmfunctions.tracer import tracer
from lmfunctions.utils import from_jsonschema, lazy_import

curdir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(curdir, "metaprompt.jinja"), "r") as f:
    default_metaprompt = f.read()

InputArgs = ParamSpec("InputArgs")
ReturnType = TypeVar("ReturnType")


class LMFunc(Base, Generic[InputArgs, ReturnType]):
    """
    A class that represents a function implemented via a language model computation (language function). Language functions are serializable objects identified by the following attributes.

    Attributes:
        name: str
            A unique identifier for the language function.
        description: str
            A brief summary of what the language function does.
        input_schema: dict
            A JSON schema that defines the expected structure and constraints
            for input data.
        output_schema: dict
            A JSON schema that defines the expected structure and constraints
            for output data.
        metaprompt: str
            A Jinja template that combines the description, input schema, and output
            schema to generate a dynamic prompt at call time.

    Methods:
        __call__(*args, examples=[], backend=None, retry_policy=None, event_manager=None, extra_args={}, **kwargs) -> ReturnType:
            Calls the language function to generate a response based on the given inputs.
            When called, the language function relies on the following components:
            - A language model backend used to perform the computation.
            - A retry policy to handle exceptions and ensure reliable execution.
            - An event manager that invokes callback functions in correspondence to events.
    """

    name: str
    description: str = ""
    input_schema: Optional[Dict] = None
    output_schema: Optional[Dict] = None
    metaprompt: str = default_metaprompt

    _input_model: Optional[Type[BaseModel]] = None
    _output_model: Optional[Type[BaseModel]] = None
    _template: Optional[Template] = None

    @staticmethod
    def to_json_str(obj: Any) -> Optional[str]:
        """
        Returns a string representation of the given object which is
        a JSON dumps for BaseModel or dict objects, and a str otherwise.
        """
        if obj is None:
            return None
        elif isinstance(obj, dict):
            return json.dumps(obj, default=str)
        elif isinstance(obj, BaseModel):
            return obj.model_dump_json()
        else:
            return str(obj)

    @property
    def input_model(self) -> Type[BaseModel]:
        if self._input_model is None:
            self._input_model = from_jsonschema(self.input_schema or {})
        return self._input_model

    @property
    def output_model(self) -> Type[BaseModel]:
        if self._output_model is None:
            self._output_model = from_jsonschema(self.output_schema or {})
        return self._output_model

    @property
    def template(self) -> Template:
        if self._template is None:
            template = Template(self.metaprompt).render(
                description=self.description,
                input_schema=self.to_json_str(self.input_schema),
                output_schema=self.to_json_str(self.output_schema),
            )
            self._template = Template(template)
        return self._template

    def __init__(
        self, func: Optional[Callable[InputArgs, ReturnType]] = None, **kwargs
    ):
        """
        Initializes the language function with the provided model fields or a Python function.

        Args:
            func (Callable[InputArgs, ReturnType], optional): A Python function used to extract the name, description, input schema, and output schema of the language function. Defaults to None.
            **kwargs: If a Python function is not provided, we initialize directly the model fields: name, description, input_schema, output_schema, metaprompt.
        """
        # If no function is provided, initialize the model fields
        if not func:
            return super().__init__(**kwargs)

        # Extract name and description
        name = func.__name__
        description = re.sub(r"\s+", " ", (func.__doc__ or "").strip())

        # Extract input and output schemas
        signature = inspect.signature(func)
        type_hints = get_type_hints(func)
        return_hint = type_hints.pop("return", None)
        fields = {
            name: (
                type_hints.get(name, str),
                ... if bool(param.empty) else param.default,
            )
            for name, param in signature.parameters.items()
        }

        # Input Schema
        if len(fields) == 0:
            input_schema = None
        else:
            first_field = next(iter(fields.items()))
            first_type_hint = first_field[1][0]
            if (
                len(fields) == 1
                and type(first_type_hint) is type(BaseModel)
                and issubclass(first_type_hint, BaseModel)
            ):
                # If the only argument is a Pydantic model, get the input schema from it
                input_schema = first_type_hint.model_json_schema()
            elif len(fields) == 1 and first_type_hint is str:
                # If the only argument is a string, input_schema is None
                input_schema = None
            else:
                # Otherwise, create a wrapper model with the input arguments as fields
                input_schema = create_model(
                    "InputWrapper", **fields
                ).model_json_schema()

        # Output Schema
        if type(return_hint) is type(BaseModel) and issubclass(return_hint, BaseModel):
            # If the return hint is a Pydantic model, get the output schema from it
            output_schema = return_hint.model_json_schema()
        elif return_hint is str and not (description):
            # If the return hint is a string and the description is empty,
            # output_schema is None. This allows to express a plain language model
            # call as a language function.
            output_schema = None
        else:
            # Otherwise, create a wrapper model with the return hint as the output type
            output_schema = create_model(
                "OutputWrapper", output=(return_hint or NoneType, ...)
            ).model_json_schema()

        super().__init__(
            name=name,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
        )

    def load(self, *args, **kwargs):
        """On load, reset the private variables."""
        super().load(*args, **kwargs)
        self._input_model = None
        self._output_model = None
        self._template = None

    def __call__(
        self,
        *args,
        examples: List = [],
        backend: Optional[LMBackend] = None,
        retry_policy: Optional[RetryPolicy] = None,
        event_manager: Optional[EventManager] = None,
        extra_args={},
        **kwargs,
    ) -> ReturnType:
        """
        Calls the language function to generate a response based on the given inputs.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
            examples (List[Tuple[Any, ReturnType]], optional): A list of examples to provide to the language model. Defaults to [].
            backend (LMBackend, optional): The language model backend to use for the computation.
            retry_policy (RetryPolicy, optional): The retry policy to use for handling exceptions.
            event_manager (EventManager, optional): The event manager to use for handling callbacks.
            extra_args (Dict, optional): Additional arguments that may be used by the callback handlers. Defaults to {}.
        """
        tracer = trace.get_tracer(__name__)
        with tracer.start_span("lmfunc.__call__") as span:
            backend = backend or default.backend
            event_manager = event_manager or default.event_manager
            retry_policy = retry_policy or default.retry_policy

            # Call Start
            event_manager(
                "call_start",
                func=self,
                args=args,
                kwargs=kwargs,
                examples=examples,
                backend=backend,
                retry_policy=retry_policy,
                event_manager=event_manager,
                extra_args=extra_args,
                tracer=tracer,
                span=span,
            )

            # Assemble all input arguments into a single input object.
            # The object can only be a Pydantic model, a dictionary, a string, or None.
            input = None
            if args or kwargs:
                first_arg = args[0] if args else next(iter(kwargs.values()), None)
                if len(args) + len(kwargs) == 1 and (
                    isinstance(first_arg, BaseModel) or isinstance(first_arg, str)
                ):
                    # If the only argument is a Pydantic model or a string, use it as input
                    input = first_arg
                else:
                    # Otherwise, create a dictionary with the input arguments
                    input = {
                        **dict(zip(self.input_model.model_fields.keys(), args)),
                        **kwargs,
                    }
            try:
                for attempt in Retrying(
                    **retry_policy.args,
                    before_sleep=lambda x: event_manager("retry", retry_call_state=x),
                ):
                    with attempt:
                        # Render Input as a string
                        input_string = self.to_json_str(input)

                        # Render Examples as strings
                        examples_string = [
                            (self.to_json_str(i), self.to_json_str(o))
                            for i, o in examples
                        ]

                        # Render the Prompt
                        prompt = self.template.render(
                            inputs=input_string,
                            examples=examples_string,
                        )

                        # Language Model Prompt Template Render Callback
                        event_manager(
                            "prompt_render",
                            func=self,
                            args=args,
                            kwargs=kwargs,
                            examples=examples,
                            backend=backend,
                            retry_policy=retry_policy,
                            event_manager=event_manager,
                            extra_args=extra_args,
                            tracer=tracer,
                            span=span,
                            ##
                            input=input,
                            attempt=attempt,
                            input_string=input_string,
                            examples_string=examples_string,
                            prompt=prompt,
                        )

                        # Language Model Call
                        response = backend.complete(prompt, self.output_schema)

                        # Parse the response
                        parsed_completion = response(
                            self.output_schema,
                            new_token_or_char_callback=lambda **kwargs: event_manager(
                                "token_or_char", span=span, **kwargs
                            ),
                        )
                        completion = response.text
                        if isinstance(parsed_completion, dict):
                            # If the completion is a dictionary, build a Pydantic object
                            output = self.output_model(**parsed_completion)
                            if self.output_model.__name__ == "OutputWrapper":
                                # If the object is a wrapper, unwrap it
                                output = getattr(
                                    output, next(iter(output.model_fields.keys()))
                                )
                        else:
                            # Otherwise, the output is the parsed completion
                            output = parsed_completion

                        # Success Callback
                        event_manager(
                            "success",
                            func=self,
                            args=args,
                            kwargs=kwargs,
                            examples=examples,
                            backend=backend,
                            retry_policy=retry_policy,
                            event_manager=event_manager,
                            extra_args=extra_args,
                            tracer=tracer,
                            span=span,
                            input=input,
                            attempt=attempt,
                            input_string=input_string,
                            examples_string=examples_string,
                            prompt=prompt,
                            ##
                            response=response,
                            parsed_completion=parsed_completion,
                            completion=completion,
                            output=output,
                        )

            except Exception as exception:
                # Exception Callback
                event_manager("exception", exception=exception, vars=locals())

            return output

    def async_handler(self):
        """
        Returns an async route handler for the language function that can be used with
        FastAPI.

        Returns:
            Callable: A FastAPI route handler for the language function.
        """

        async def handler(input=None):
            return self(**input) if isinstance(input, dict) else self(input)

        handler.__annotations__["input"] = self.input_model
        if self.output_model.__name__ == "OutputWrapper":
            handler.__annotations__["return"] = next(
                iter(self.output_model.model_fields.values())
            ).annotation
        else:
            handler.__annotations__["return"] = self.output_model
        return handler

    def fastapi_app(self, fast_api_params: Dict = {}):
        """
        Creates a FastAPI application with the specified parameters and registers
        a POST route for the current instance.

        Args:
            fast_api_params (Dict, optional): Additional parameters to be passed to the
            FastAPI application. Defaults to {}.

        Returns:
            FastAPI: The created FastAPI application.
        """
        lazy_import("fastapi")
        import fastapi

        app = fastapi.FastAPI(**fast_api_params)
        app.post(
            f"/{self.name}",
            name=self.name,
            description=self.description,
        )(self.async_handler())

        return app

    def serve(self, fast_api_params={}, uvicorn_params={}):
        """
        Serves the lmfunc using FastAPI and Uvicorn.

        Args:
            fast_api_params (dict): Parameters to be passed to the FastAPI application.
            uvicorn_params (dict): Parameters to be passed to the Uvicorn server.

        Returns:
            None
        """

        def start_uvicorn_server(app, **kwargs):
            lazy_import("uvicorn")
            import uvicorn

            return uvicorn.run(app, **kwargs)

        app = self.fastapi_app(fast_api_params)
        if app:
            return start_uvicorn_server(app, **uvicorn_params)


def lmdef(func: Callable[InputArgs, ReturnType]):
    """
    Decorator that creates an LMFunc instance from a regular Python function.
    """
    return LMFunc[InputArgs, ReturnType](func)



from typing import Literal

from lmfunctions.handler import Handler


class ExceptionRaiseHandler(Handler):
    """Event handler that raises an exception extracted from the event variables."""

    name: Literal["exceptionraise"] = "exceptionraise"
    varname: str = "exception"

    def __call__(self, **kwargs):
        raise kwargs[self.varname]



from typing import Dict, Literal

from lmfunctions.handler import Handler


class OtelEventHandler(Handler):
    """Callback handler that sends events to an OpenTelemetry span."""

    name: Literal["otelevent"] = "otelevent"
    span_name: str = "span"
    event_name: str = "event"
    attributes_vars: Dict[str, str] = {}

    def __call__(self, **kwargs):
        span = kwargs.get(self.span_name, None)
        attributes = {
            key: kwargs.get(varname, None)
            for key, varname in self.attributes_vars.items()
        }
        if span:
            span.add_event(self.event_name, attributes=attributes)



from typing import Optional

from opentelemetry import trace
from opentelemetry.sdk.trace import SpanLimits, TracerProvider
from pandas import DataFrame
from rich import print

from lmfunctions.eventmanager import EventManager
from lmfunctions.handler import ExceptionRaiseHandler, OtelEventHandler


def time_diff(df, event_type1, event_type2):
    return (
        df.loc[event_type1]["timestamp"].min() - df.loc[event_type2]["timestamp"].max()
    ) / 10**9


def print_stats(**kwargs):
    prompt = kwargs["prompt"]
    completion = kwargs["completion"]
    span = kwargs["span"]
    events = DataFrame.from_records(
        [
            {"name": e.name, "timestamp": e.timestamp, "attributes": e.attributes}
            for e in span.events
        ],
        index="name",
    )
    prompt_length = len(prompt)
    completion_length = len(completion)
    time_to_prompt_render = time_diff(events, "prompt_render", "call_start")
    time_to_first_token = time_diff(events, "token_or_char", "call_start")
    time_to_first_token_lm_only = time_diff(events, "token_or_char", "prompt_render")
    lm_call_time = time_diff(events, "success", "prompt_render")
    total_time = time_diff(events, "success", "call_start")
    stats = {
        "Prompt Characters": prompt_length,
        "Completion Characters": completion_length,
        "Time to Prompt Render": time_to_prompt_render,
        "Time to First Token": time_to_first_token,
        "Time to First Token (LM Only)": time_to_first_token_lm_only,
        "LM Call Time": lm_call_time,
        "Total Time": total_time,
        "Completion Characters per Second": completion_length / total_time,
        "Completion Characters / LM Call Time": completion_length / lm_call_time,
        "Prompt Characters / LM Call Time": prompt_length / lm_call_time,
    }
    print(stats)


def timeEvents(tracer_provider: Optional[TracerProvider] = None) -> EventManager:
    """ """

    if tracer_provider is None:
        tracer_provider = trace.get_tracer_provider()
        if not isinstance(tracer_provider, TracerProvider):
            # Set the tracer provider with custom limits
            tracer_provider = TracerProvider(span_limits=SpanLimits(max_events=1000000))
            trace.set_tracer_provider(tracer_provider)

    return EventManager(
        handlers={
            "call_start": [OtelEventHandler(event_name="call_start")],
            "prompt_render": [OtelEventHandler(event_name="prompt_render")],
            "token_or_char": [
                OtelEventHandler(
                    event_name="token_or_char",
                    attributes_vars={"token_or_char": "token_or_char"},
                )
            ],
            "success": [OtelEventHandler(event_name="success"), print_stats],
            "exception": [
                OtelEventHandler(event_name="exception"),
                ExceptionRaiseHandler(varname="exception"),
            ],
        }
    )



from enum import Enum

import yaml
from tenacity import (
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
    wait_fixed,
    wait_none,
    wait_random,
)

from lmfunctions.base import Base


class StopType(str, Enum):
    never = "never"
    after_attempt = "after_attempt"
    after_delay = "after_delay"


class WaitType(str, Enum):
    none = "none"
    fixed = "fixed"
    random = "random"
    exponential = "exponential"


class RetryPolicy(Base):
    """
    A retry policy in a serializable format. It generates retry arguments
    for the tenacity library. See https://tenacity.readthedocs.io/en/latest/.
    """

    stop: StopType = StopType.after_attempt
    wait: WaitType = WaitType.none
    stop_max_attempt: int = 3
    stop_max_delay: int = 1
    wait_fixed: int = 1
    wait_random_min: int = 1
    wait_random_max: int = 2
    wait_exponential_min: int = 1
    wait_exponential_max: int = 2
    wait_exponential_multiplier: float = 2.0
    reraise: bool = True

    @property
    def args(self) -> dict:
        """
        Get the retry arguments based on the configured retry policy.

        Returns:
            dict: The retry arguments to be used for retrying.
        """
        retry_args = {}
        if self.stop == StopType.after_attempt:
            retry_args["stop"] = stop_after_attempt(self.stop_max_attempt)
        elif self.stop == StopType.after_delay:
            retry_args["stop"] = stop_after_delay(self.stop_max_delay)
        if self.wait == WaitType.none:
            retry_args["wait"] = wait_none()
        elif self.wait == WaitType.fixed:
            retry_args["wait"] = wait_fixed(self.wait_fixed)
        elif self.wait == WaitType.random:
            retry_args["wait"] = wait_random(self.wait_random_min, self.wait_random_max)
        elif self.wait == WaitType.exponential:
            retry_args["wait"] = wait_exponential(
                min=self.wait_exponential_min,
                max=self.wait_exponential_max,
                multiplier=self.wait_exponential_multiplier,
            )
        retry_args["reraise"] = self.reraise
        return retry_args


# Workaround for issue with YAML serialization of enums
# See https://github.com/yaml/pyyaml/issues/722
for enumtype in [StopType, WaitType]:
    yaml.SafeDumper.add_representer(
        enumtype,  # type: ignore
        yaml.representer.SafeRepresenter.represent_str,
    )



import importlib.util
import json
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Type

from datamodel_code_generator import DataModelType, InputFileType, generate
from pydantic import BaseModel


def from_jsonschema(schema: Dict) -> Type[BaseModel]:
    """Generate a Pydantic Model from a json schema.

    Args:
    schema: Source json schema to create Pydantic model from

    Returns:
    The newly created and loaded Pydantic class
    """
    # Ref: https://github.com/koxudaxi/datamodel-code-generator/issues/278
    class_name = schema.get("title", "Model")
    json_schema = json.dumps(schema)
    with TemporaryDirectory() as temporary_directory_name:
        temporary_directory = Path(temporary_directory_name)
        temporary_file_path = Path(temporary_directory / "tempmodel.py")
        generate(
            json_schema,
            input_file_type=InputFileType.JsonSchema,
            class_name=class_name,
            output=temporary_file_path,
            output_model_type=DataModelType.PydanticV2BaseModel,
        )
        spec = importlib.util.spec_from_file_location(
            "models", str(temporary_file_path)
        )
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module
            spec.loader.exec_module(module)
            return getattr(module, class_name)
        raise ImportError("Failed to import generated model")  # pragma: no cover



from lmfunctions.eventmanager import EventManager
from lmfunctions.handler import PanelPrintHandler, PrintHandler


def panelPrint() -> EventManager:
    """
    EventManager preset that prints selected internal variables to the console
    using panels.
    """
    return EventManager(
        handlers={
            "prompt_render": [
                PanelPrintHandler(variables=[dict(name="prompt", title="Prompt")])
            ],
            "token_or_char": [
                PrintHandler(varnames=["token_or_char"], end="", flush=True)
            ],
            "retry": [
                PanelPrintHandler(
                    variables=[dict(name="retry_call_state", title="Retry")]
                )
            ],
            "exception": [
                PanelPrintHandler(variables=[dict(name="exception", title="Exception")])
            ],
            "success": [
                PanelPrintHandler(
                    variables=[
                        dict(name="completion", title="Completion"),
                        dict(name="parsed_completion", title="Parsed Completion"),
                        dict(name="output", title="Output"),
                    ]
                )
            ],
        }
    )



from .litellm import LiteLLMBackend
from .llamacpp import LlamaCppBackend
from .transformers import TransformersBackend

__all__ = [
    "LiteLLMBackend",
    "LlamaCppBackend",
    "TransformersBackend",
]



from typing import Any, Dict, Optional

from pydantic import BaseModel, ConfigDict

from lmfunctions.objectstore import Hub, ObjectStore
from lmfunctions.utils import dumps, loadf, loads, panelprint

hub = Hub()


class Base(BaseModel):
    """
    A customized Pydantic model with convenience methods for serialization,
    deserialization, storage. Enforces validation of the model's data on
    assignment to fields.
    """

    model_config = ConfigDict(
        protected_namespaces=(), validate_assignment=True, arbitrary_types_allowed=True
    )

    def dump(self, **kwargs) -> Dict[str, Any]:
        """
        Returns a dump of the model.

        This method returns a dump of the current state of the model.
        It can be used for debugging or saving the model's state.

        Returns:
            str: A string representation of the model's dump.
        """
        return self.model_dump(**kwargs)

    def dumps(self, format: str = "yaml", **kwargs) -> str:
        """
        Serialize the data model object to a string representation.

        Args:
            format (str, optional): The serialization format. Defaults to "yaml".

        Returns:
            str: The serialized string representation of the data model object.
        """
        return dumps(self.dump(**kwargs), format=format)

    def push(self, path: str, objectstore: ObjectStore = hub, **kwargs) -> None:
        """
        Pushes the serialized object to an object store.

        Args:
        - path (str): The path to save the object to in the object store.
        """
        return objectstore.push(path, self.dumps(), **kwargs)

    def load(self, data: Optional[Dict] = None, **kwargs) -> None:
        """
        Loads data into the object, checking that it conforms to the model.

        Args:
            data (dict): The data to be loaded.

        Returns:
            None
        """
        new_data = self.dump() | (data or kwargs)
        self.__dict__.update(self.model_validate(new_data))

    def loads(self, data, format: str = "yaml") -> None:
        """
        Deserializes a string in the given format to an object.

        Args:
        - data (str): The string to deserialize.
        - format (str): The format to deserialize from.
        """
        self.load(loads(data, format=format))

    def loadf(self, url: str, **kwargs) -> None:
        """
        Loads the object from a generalized URL.

        The format is inferred from the file extension.

        Args:
        - url (str): The URL to load the configuration from.
        """
        self.load(loadf(url, **kwargs))

    def pull(self, path: str, objectstore: ObjectStore = hub, **kwargs) -> None:
        """
        Pulls the object from the object store.

        Args:
        - path (str): The path to the object in the object store.
        """
        self.load(objectstore.pull(path, **kwargs))

    @classmethod
    def from_string(cls, data, format: str = "yaml"):
        """
        Creates an instance from a string.

        Args:
        - data (str): The string to create the instance from.
        - format (str): The format of the string.

        Returns:
        - An instance created from the string.
        """
        return cls(**loads(data, format=format))

    @classmethod
    def from_file(cls, url: str, **kwargs):
        """
        Creates an instance by reading a file-like object from a generalized URL.

        Args:
        - url (str): The URL to create the instance from.

        Returns:
        - An instance created from the URL.
        """
        return cls(**loadf(url, **kwargs))

    @classmethod
    def from_store(cls, path: str, objectstore: ObjectStore = hub, **kwargs):
        """
        Creates an instance by pulling it from an object store.

        Args:
        - path (str): The path to the object in the object store.

        Returns:
        - An instance created from the object store.
        """
        return cls(**objectstore.pull(path, **kwargs))

    def info(self) -> None:
        """
        Prints information about the object.

        This method uses the `panel_print` function to print information about
        the object. The title of the printed information is set to the name of
        the class.
        """
        panelprint(self, title=self.__class__.__name__)



from functools import partial
from typing import Any, Callable, List, Optional

from rich import print

from lmfunctions.default import default
from lmfunctions.lmbackend.lmbackend import LMBackend


def multiline_input(terminators=tuple()):
    """
    Simple input function that allows entering multiple lines
    """
    lines = []
    prompt = ">>> "
    while True:
        line = input(prompt)
        lines.append(line)
        if line in terminators:
            break
        prompt = "... "
    return "\n".join(lines)


def default_setup_callback(backend: LMBackend):
    backend.complete("warmup json", schema={"type": "null"})
    print(
        f"Backend: {backend.name}\nModel: {backend.model}\n"
        "Type /exit to exit.\n"
        "Type /history to see all messages\n"
        "Type /clear to clear the chat history\n"
        "Interrupt generation with Ctrl+C (streaming-mode only)\n"
    )


def chat(
    backend: Optional[LMBackend] = None,
    setup_callback: Optional[Callable] = default_setup_callback,
    input_callback: Callable[[], str] = partial(
        multiline_input, terminators=("", "\n", "/exit", "/clear", "/history")
    ),
    new_token_or_char_callback: Callable[[Any], None] = lambda **kwargs: print(
        kwargs["token_or_char"], end="", flush=True
    ),
):
    """
    A minimal chat loop that interacts with the default runtime language model.
    """
    if backend is None:
        backend = default.backend
    if setup_callback:
        setup_callback(backend)
    messages: List = []
    while True:
        user_message = input_callback()
        if not user_message:
            continue
        if user_message in ("/exit", "/clear"):
            if user_message == "/exit":
                return
            else:
                messages = []
        elif user_message == "/history":
            print(messages)
        else:
            messages.append({"role": "user", "content": user_message})
            try:
                response = backend.chat_complete(messages)
                response(new_token_or_char_callback=new_token_or_char_callback)
            except KeyboardInterrupt:  # pragma: no cover
                pass  # pragma: no cover
            finally:
                print()
                messages.append({"role": "assistant", "content": response()})



from .handler import Handler
from .handlers import (
    ExceptionRaiseHandler,
    LoggingHandler,
    OtelEventHandler,
    PanelPrintHandler,
    PrintHandler,
)

__all__ = [
    "PanelPrintHandler",
    "LoggingHandler",
    "PrintHandler",
    "Handler",
    "ExceptionRaiseHandler",
    "OtelEventHandler",
]



import gc
from typing import Any, Dict, List, Literal, Optional

from lmformatenforcer import JsonSchemaParser
from pydantic import model_validator

from lmfunctions.lmbackend.lmbackend import LMBackend
from lmfunctions.lmresponse import LMResponse
from lmfunctions.utils import cuda_check, lazy_import


class TransformersBackend(LMBackend):
    name: Literal["transformers"] = "transformers"
    model: str = "Qwen/Qwen2-0.5B"
    config: str | None = None
    feature_extractor: str | None = None
    image_processor: str | None = None
    framework: str | None = None
    revision: str | None = None
    use_fast: bool = True
    token: str | bool | None = None
    device: int | str | None = None
    device_map: str | None = None
    trust_remote_code: bool = False
    model_kwargs: Dict[str, Any] = {}
    pipeline_class: Any | None = None
    generation: Dict[str, Any] = {}

    _pipeline: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.device is None:
            gpu_info = cuda_check()
            if not gpu_info["cuda_available"]:
                self.device = "cpu"
            elif gpu_info["num_gpus"] > 1 and self.device_map is None:
                try:
                    __import__("accelerate")

                    self.device_map = "auto"
                except ImportError:
                    self.device = "cuda"
            else:
                self.device = "cuda"

    @property
    def pipeline(self):
        if self._pipeline is None:
            if lazy_import("transformers", install_packages=["transformers['torch']"]):
                import transformers

                tokenizer = transformers.AutoTokenizer.from_pretrained(self.model)
                self._pipeline = transformers.pipeline(
                    task="text-generation",
                    tokenizer=tokenizer,
                    **self.model_dump(exclude={"name", "generation"})
                )
                self._pipeline.model.generation_config.pad_token_id = (
                    tokenizer.eos_token_id
                )

            else:
                raise ImportError(
                    "The package 'transformers' is required"
                )  # pragma: no cover
        return self._pipeline

    def _unload(self):
        self._pipeline = None
        gc.collect()
        try:
            import torch

            torch.cuda.empty_cache()
        except ImportError:  # pragma: no cover
            pass  # pragma: no cover
        return self

    @model_validator(mode="after")
    def unload(self):
        if self._pipeline:
            self._unload()
        return self

    @property
    def prefix_fn(self):
        from lmformatenforcer.integrations.transformers import (
            build_transformers_prefix_allowed_tokens_fn,
        )

        return build_transformers_prefix_allowed_tokens_fn

    def complete(
        self, prompt: str = "", schema: Optional[Dict] = None, **kwargs
    ) -> LMResponse:
        messages = [{"role": "user", "content": prompt}]
        pipeline = self.pipeline
        if schema and pipeline.tokenizer:
            prefix_function = self.prefix_fn(
                pipeline.tokenizer, JsonSchemaParser(schema)
            )
            return self.chat_complete(
                messages, prefix_allowed_tokens_fn=prefix_function, **kwargs
            )
        else:
            return self.chat_complete(messages, **kwargs)

    def chat_complete(self, messages: List, **kwargs) -> LMResponse:
        generation_config = self.pipeline.model.generation_config
        if (
            generation_config.max_length == 20
            and generation_config.max_new_tokens is None
            and "max_new_tokens" not in self.generation
            and "max_length" not in self.generation
        ):
            self.generation["max_new_tokens"] = 4096

        params = self.generation | kwargs
        response = self.pipeline(messages, **params)
        return LMResponse(response[0]["generated_text"][-1].get("content", ""))



import logging
from typing import List, Literal, Optional

from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import pretty_repr

from lmfunctions.handler.handler import Handler


class LoggingHandler(Handler):
    """
    Callback handler that log events to a logger.

    Attributes:
        name (Literal["logging"]): The name of the handler.
        logger_name (Optional[str]): The name of the logger to use.
        If not provided, a default logger will be used.
        log_level (int): The log level to use for logging events.
        log_time_format (str): The format string for the log timestamp.
        log_file (Optional[str]): The path to the log file.
        If provided, file logging will be used.
        file_logging_format (str): The format string for file logging.
        message (str): The message to include in the log event.
        log_keys (List[str]): The keys of the variables to include in the log event.

    """

    name: Literal["logging"] = "logging"
    logger_name: Optional[str] = "DEFAULT_LOGGER"
    log_level: int = logging.INFO
    log_time_format: str = "%Y-%m-%d %H:%M:%S"
    log_file: Optional[str] = None
    file_logging_format: str = (
        "%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s"
    )
    message: str = "Logging event"
    log_keys: List[str] = []

    _logger: Optional[logging.Logger] = None

    @property
    def logger(self):
        if not self._logger:
            logger = logging.getLogger(self.logger_name)
            logger.setLevel(self.log_level)
            if self.log_file:
                # Create a logger for file logging
                handler = logging.FileHandler(self.log_file)
                handler.setFormatter(
                    logging.Formatter(
                        fmt=self.file_logging_format,
                        datefmt=self.log_time_format,
                    )
                )
            else:
                # Create a logger for stdout with rich handler
                handler = RichHandler(
                    console=Console(),
                    rich_tracebacks=True,
                    log_time_format=self.log_time_format,
                )
            logger.handlers = [handler]
            self._logger = logger
        return self._logger

    def __call__(self, **kwargs):
        filtered_vars = {key: kwargs[key] for key in self.log_keys if key in kwargs}
        pretty_vars = pretty_repr(filtered_vars)
        log_message = f"{self.message}\n{pretty_vars}"
        self.logger.log(self.log_level, log_message)



from typing import List, Literal

from lmfunctions.handler import Handler


class PrintHandler(Handler):
    """Callback handler that prints the specified variables to the console."""

    name: Literal["print"] = "print"
    sep: str | None = " "
    end: str | None = "\n"
    flush: bool = False
    varnames: List[str] = []

    def __call__(self, **kwargs):
        print(
            *(kwargs.get(variable, None) for variable in self.varnames),
            sep=self.sep,
            end=self.end,
            flush=self.flush,
        )



from typing import Dict, List, Literal

from lmfunctions.handler import Handler
from lmfunctions.utils import panelprint


class PanelPrintHandler(Handler):
    """Callback handler that prints the specified variables to the console using titled panels."""

    name: Literal["panelprint"] = "panelprint"
    variables: List[Dict] = []

    def __call__(self, **kwargs):
        for variable in self.variables:
            name = variable.get("name", None)
            title = variable.get("title", name)
            panelprint(kwargs.get(name, None), title=title)



from .backends import (
    LiteLLMBackend,
    LlamaCppBackend,
    TransformersBackend,
)
from .lmbackend import LMBackend

__all__ = [
    "LMBackend",
    "LiteLLMBackend",
    "LlamaCppBackend",
    "TransformersBackend",
]



from abc import abstractmethod
from typing import Dict


class ObjectStore:
    """Generic interface for an object store."""

    @abstractmethod
    def push(self, obj, path, **kwargs):
        pass  # pragma: no cover

    @abstractmethod
    def pull(self, path, **kwargs) -> Dict:
        pass  # pragma: no cover



from .exceptionraise import ExceptionRaiseHandler
from .logging import LoggingHandler
from .otelevent import OtelEventHandler
from .panelprint import PanelPrintHandler
from .print import PrintHandler

__all__ = [
    "LoggingHandler",
    "PanelPrintHandler",
    "PrintHandler",
    "ExceptionRaiseHandler",
    "OtelEventHandler",
]



from typing import Any, get_args

from rich import print
from rich.console import RenderableType
from rich.panel import Panel
from rich.pretty import Pretty


def panelprint(_object: Any, title: str = ""):

    print(
        "\n",
        Panel(
            (
                _object
                if isinstance(_object, get_args(RenderableType))
                else Pretty(_object)
            ),
            title=title,
        ),
    )



import subprocess
from importlib import import_module
from types import ModuleType


def pip_install(packages):
    for package in packages:
        response = input(
            f"Package '{package}' is needed. Do you want to run "
            f"``pip install {package}''? (y/n)"
        )
        if response.lower() == "y":
            subprocess.run(["pip", "install", package])
            print(f"Package '{package}' installed successfully.")
        else:
            return False
    return True


def lazy_import(name, package=None, install_packages=None) -> ModuleType | None:
    """
    Lazily imports a module and handles the case when the module is not found.

    Args:
        module_name (str): The name of the module to import..

    Returns:
        ModuleType | None: The imported module if found, otherwise None.

    Raises:
        None
    """

    try:
        return import_module(name, package=package)
    except ImportError:
        install_packages = install_packages or [name]
        if pip_install(install_packages):
            return import_module(name, package=package)
        raise ImportError(f"The following packages are required: '{install_packages}'")



from .consolerich import consoleRich
from .filelog import fileLog
from .panelprint import panelPrint
from .timeevents import timeEvents
from .tokenstream import tokenStream

__all__ = ["consoleRich", "fileLog", "panelPrint", "tokenStream", "timeEvents"]



import logging

from lmfunctions.eventmanager import EventManager
from lmfunctions.handler import LoggingHandler


def consoleRich() -> EventManager:
    """EventManager preset that logs selected internal variables to the console
    using the rich library."""

    return EventManager(
        handlers={
            "call_start": [
                LoggingHandler(
                    message="Call started with vars:",
                    log_keys=["self", "input", "examples", "runtime", "kwargs"],
                )
            ],
            "prompt_render": [
                LoggingHandler(
                    message="Prompt rendered with vars:",
                    log_keys=["input_string", "examples_string", "prompt"],
                )
            ],
            "token_or_char": [
                LoggingHandler(
                    message="New token/char processed:",
                    log_keys=["token_or_char"],
                )
            ],
            "retry": [
                LoggingHandler(
                    message="Retrying with vars:", log_keys=["retry_call_state"]
                )
            ],
            "exception": [
                LoggingHandler(
                    log_level=logging.ERROR,
                    message="Exception occurred:",
                    log_keys=["exception"],
                )
            ],
            "success": [
                LoggingHandler(
                    message="Process succeeded with vars:",
                    log_keys=["input", "completion", "parsed_completion", "output"],
                )
            ],
        }
    )



import json
import os
from typing import Any, Dict

import fsspec
import yaml


def dumps(data, format="yaml"):
    """
    Serialize the given data into a string representation.

    Args:
        data: The data to be serialized.
        format (str): The format in which to serialize the data.
        Supported formats are "yaml" and "json".

    Returns:
        str: The serialized data as a string.

    Raises:
        ValueError: If the specified format is not supported.
    """
    if format == "yaml":
        return yaml.safe_dump(data, default_flow_style=False)
    elif format == "json":
        return json.dumps(data)
    else:
        raise ValueError("Unsupported format")


def loads(data, format="yaml") -> Dict[str, Any]:
    """
    Load a dictionary from a string representation using the specified format.

    Args:
        data (str): The string representation of the data.
        format (str, optional): The format of the data. Defaults to "yaml".

    Returns:
        object: The loaded data.

    Raises:
        ValueError: If the specified format is unsupported.
    """
    if format == "yaml":
        return yaml.safe_load(data)
    elif format == "json":
        return json.loads(data)
    else:
        raise ValueError("Unsupported format")


def loadf(url: str, **kwargs) -> Dict[str, Any]:
    """
    Load a file from the given generalized URL and return its contents.
    Determines the format of the file from its extension.

    Parameters:
    url (str): The generalized URL of the file to load.
    **kwargs: Additional keyword arguments to pass to the `fsspec.open` function.

    Returns:
    The contents of the file.

    """
    extension = os.path.splitext(url)[1][1:]
    with fsspec.open(url, **kwargs) as file:
        return loads(file, format=extension)



from .dictutils import dumps, loadf, loads
from .importutils import lazy_import, pip_install
from .panelprint import panelprint
from .pydantic import from_jsonschema
from .cuda_check import cuda_check

__all__ = [
    "panelprint",
    "lazy_import",
    "from_jsonschema",
    "dumps",
    "loads",
    "loadf",
    "pip_install",
    "cuda_check",
]
